/******************************************************************************
Module:  ServiceCtrl.h
Notices: Copyright (c) 2000 Jeffrey Richter
Purpose: This class wraps the control of a Service application.
         See Appendix B.
******************************************************************************/

#pragma once   // Include this header file once per compilation unit


///////////////////////////////////////////////////////////////////////////////

#ifndef _UNICODE
#error This module must be compiled natively using Unicode.
#endif

#define WINVER  0x0500  
#define _WIN32_WINNT  0x0500 //Usefull!!!!!
#define _WIN32_IE 0x0600

//#include "Winsvc.h"
///////////////////////////////////////////////////////////////////////////////

#include "CmnHdr.h"              // See Appendix A.
#include <malloc.h>

#include "AutoBuf.h"    // See Appendix B.

#define SECINFO_IMPL
#include "SecInfo.h"    // See Appendix B.

#include "SCMCTRL.H"
///////////////////////////////////////////////////////////////////////////////


class CServiceCtrl {
public:
   CServiceCtrl();
   CServiceCtrl(CSCMCtrl& scm, BOOL fInternalName, 
      PCTSTR pszName, DWORD dwDesiredAccess);
   ~CServiceCtrl();

   BOOL InstallAndOpen(CSCMCtrl& scm, PCTSTR pszInternalName, 
      PCTSTR pszDisplayName, PCTSTR pszDescription, DWORD dwServiceType, 
      DWORD dwStartType, DWORD dwErrorControl, PCTSTR pszPathname, 
      PCTSTR pszLoadOrderGroup, PCTSTR pszDependencies, 
      PCTSTR pszUserName, PCTSTR pszUserPswd, DWORD dwDesiredAccess = 0);

   BOOL Open(CSCMCtrl& scm, BOOL fInternalName, 
      PCTSTR pszName, DWORD dwDesiredAccess);
   BOOL OpenOK() const { return(m_h != NULL); }
   operator SC_HANDLE() const { return(m_h); }
   BOOL Delete();
   void ForceClose() { Reconstruct(); }

   BOOL Start(DWORD nNumArgs, PCTSTR *pArgs, BOOL fWait = FALSE, 
      SERVICE_STATUS* pss = NULL, DWORD dwMilliseconds = INFINITE);

   BOOL Start(PCTSTR pszArgs, BOOL fWait = FALSE, 
      SERVICE_STATUS* pss = NULL, DWORD dwMilliseconds = INFINITE);

   BOOL Control(DWORD dwControl, BOOL fWait = FALSE, 
      SERVICE_STATUS* pss = NULL, DWORD dwMilliseconds = INFINITE);

   BOOL WaitForState(DWORD dwDesiredState, 
      SERVICE_STATUS* pss = NULL, DWORD dwMilliseconds = INFINITE);

   BOOL QueryStatus(SERVICE_STATUS* pss);
   BOOL QueryStatus(SERVICE_STATUS_PROCESS* pssp);

   const QUERY_SERVICE_CONFIG* QueryConfig();
   const PCTSTR QueryDescription();
   BOOL ChangeConfig(PCTSTR pszDisplayName, PCTSTR pszDescription, 
      DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, 
      PCTSTR pBinaryPathName, PCTSTR pLoadOrderGroup, PCTSTR pDependencies, 
      PCTSTR pszUserName, PCTSTR pszUserPswd);

   const SERVICE_FAILURE_ACTIONS* QueryFailureActions();
   BOOL _cdecl ChangeFailureActions(DWORD dwResetPeriod, 
      PCTSTR pszRebootMsg, PCTSTR pszCommand, 
      SC_ACTION_TYPE Type, DWORD dwDelay, ...);

   BOOL CreateDependencySnapshot(DWORD dwServiceState);
   int GetDependencySnapshotNum();
   const ENUM_SERVICE_STATUS* GetDependencySnapshotEntry(int nIndex);

   void EditSecurity(HWND hwnd);
   SC_HANDLE GetSCHandle(){ return m_h; }
private:
   void Reconstruct(BOOL fFirstTime = FALSE);

private:
   SC_HANDLE m_h;
   union {
      BYTE bRaw[1024 + sizeof(PTSTR)];
      SERVICE_DESCRIPTION cooked;
   } m_sd;

   CAutoBuf<QUERY_SERVICE_CONFIG> m_pServiceConfig;
   CAutoBuf<ENUM_SERVICE_STATUS> m_pDependencySnapshot;
   DWORD m_nDependencySnapshotNum;

   union {
      BYTE bRaw[1024];
      SERVICE_FAILURE_ACTIONS cooked;
   } m_sfa;

private:
   class CServiceSecInfo : public CSecInfo {
   public:
      CServiceSecInfo(CServiceCtrl* pServiceCtrl, PTSTR pszMachine = NULL);

   private:
      HRESULT WINAPI GetObjectInformation(PSI_OBJECT_INFO pObjInfo);
      HRESULT WINAPI GetAccessRights(const GUID* pguidObjectType,
         DWORD dwFlags, PSI_ACCESS* ppAccess, ULONG* pcAccesses, 
         ULONG* piDefaultAccess);
      HRESULT WINAPI GetSecurity(SECURITY_INFORMATION RequestedInformation,
         PSECURITY_DESCRIPTOR* ppSecurityDescriptor, BOOL fDefault);
      HRESULT WINAPI SetSecurity(SECURITY_INFORMATION SecurityInformation,
         PSECURITY_DESCRIPTOR psd);
      HRESULT WINAPI MapGeneric(const GUID *pguidObjectType, 
         UCHAR *pAceFlags, ACCESS_MASK *pMask);

   private:
      CServiceCtrl* m_pServiceCtrl;
      PTSTR         m_pszMachine;
   };
};


///////////////////////////////////////////////////////////////////////////////


inline PCTSTR NullStringToNull(PCTSTR psz) {
   if ((psz != NULL) && (psz[0] == 0)) psz = NULL;
   return(psz);
}


///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////


//#endif   // SERVICECTRL_IMPL

///////////////////////////////// End of File /////////////////////////////////
