/******************************************************************************
Module:  SCMCtrl.h
Notices: Copyright (c) 2000 Jeffrey Richter
Purpose: This class wraps the control of a Service Control Manager.
         See Appendix B.
******************************************************************************/

#pragma once   // Include this header file once per compilation unit


///////////////////////////////////////////////////////////////////////////////
#define WINVER  0x0500  
#define _WIN32_WINNT  0x0500 //Usefull!!!!!
#define _WIN32_IE 0x0600

#include "Winsvc.h"
#include "CmnHdr.h"              // See Appendix A.
#include "AutoBuf.h"    // See Appendix B.


///////////////////////////////////////////////////////////////////////////////


class CSCMCtrl {
public:
   CSCMCtrl(DWORD dwDesiredAccess = SC_MANAGER_CONNECT, 
      PCTSTR pszMachine = NULL);
   ~CSCMCtrl();

   BOOL Open(DWORD dwDesiredAccess = SC_MANAGER_CONNECT, 
      PCTSTR pszMachine = NULL);
   BOOL OpenOK() const { return(m_h != NULL); }
   operator SC_HANDLE() const { return(m_h); }
   void ForceClose() { Reconstruct(); }

   BOOL LockDatabase(BOOL fLock = TRUE);
   const PCTSTR QueryLockOwner(PDWORD pdwLockDuration = NULL);

   const PCTSTR GetInternalName(PCTSTR pszDisplayName);
   const PCTSTR GetDisplayName(PCTSTR pszInternalName);

   BOOL CreateStatusSnapshot(DWORD dwServiceState);
   int  GetStatusSnapshotNum();
   const ENUM_SERVICE_STATUS* GetStatusSnapshotEntry(int nIndex);

private:
   void Reconstruct(BOOL fFirstTime = FALSE);

private:
   SC_HANDLE m_h;
   TCHAR m_szMachine[_MAX_PATH];
   
   SC_LOCK m_scLock;
   union {
      BYTE bRaw[1024];
      QUERY_SERVICE_LOCK_STATUS cooked;
   } m_qsls;

   TCHAR m_szInternalName[1024];
   TCHAR m_szDisplayName[1024];

   DWORD m_nStatusSnapshotNum;
   CAutoBuf<ENUM_SERVICE_STATUS> m_pStatusSnapshot;
};


///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////


//#endif   // SCMCTRL_IMPL



///////////////////////////////// End of File /////////////////////////////////
