/******************************************************************************
Module:  SecInfo.h
Notices: Copyright (c) 2000 Jeffrey Richter
Purpose: This class wraps the ISecurityInformation interface, which is used 
         in calls to the EditSecurity function.
         See Appendix B.
******************************************************************************/

//#pragma once   // Include this header file once per compilation unit
#ifndef SECURITY_INFO
#define SECURITY_INFO

///////////////////////////////////////////////////////////////////////////////

#define WINVER  0x0500  
#define _WIN32_WINNT  0x0500 //Usefull!!!!!
#define _WIN32_IE 0x0600

#include "CmnHdr.h"              // See Appendix A.
#include <aclapi.h>
#include <aclui.h>


///////////////////////////////////////////////////////////////////////////////

class CSecInfo: public ISecurityInformation {
public:
   CSecInfo() { m_nRef = 1; m_fMod = FALSE; }
   BOOL IsModified() { return(m_fMod); }

protected:
   virtual ~CSecInfo() {}

protected:
   void Modified() { m_fMod = TRUE; }
   static GUID m_guidNULL;
   static SI_ACCESS m_siAccessAllRights[];

private:
   ULONG m_nRef; 
   BOOL  m_fMod;    

public:
   HRESULT WINAPI QueryInterface(REFIID riid, PVOID* ppvObj);
   ULONG WINAPI AddRef();
   ULONG WINAPI Release();
   HRESULT UseStandardAccessRights(const GUID* pguidObjectType, DWORD dwFlags,
      PSI_ACCESS* ppAccess, ULONG* pcAccesses, ULONG* piDefaultAccess);

protected:
   HRESULT WINAPI GetObjectInformation(PSI_OBJECT_INFO pObjectInfo) = 0;
   HRESULT WINAPI GetSecurity(SECURITY_INFORMATION RequestedInformation,
      PSECURITY_DESCRIPTOR* ppSecurityDescriptor, BOOL fDefault) = 0;
   HRESULT WINAPI SetSecurity(SECURITY_INFORMATION SecurityInformation,
      PSECURITY_DESCRIPTOR pSecurityDescriptor) = 0;
   HRESULT WINAPI GetAccessRights(const GUID* pguidObjectType,
      DWORD dwFlags, // SI_EDIT_AUDITS, SI_EDIT_PROPERTIES
      PSI_ACCESS *ppAccess, ULONG *pcAccesses, ULONG *piDefaultAccess) = 0;
   HRESULT WINAPI MapGeneric(const GUID *pguidObjectType,
      UCHAR *pAceFlags, ACCESS_MASK *pMask) = 0;
   HRESULT WINAPI GetInheritTypes(PSI_INHERIT_TYPE* ppInheritTypes, 
      ULONG *pcInheritTypes);
   HRESULT WINAPI PropertySheetPageCallback(HWND hwnd, UINT uMsg, 
      SI_PAGE_TYPE uPage);

   PSECURITY_DESCRIPTOR LocalAllocSDCopy(PSECURITY_DESCRIPTOR psd);
};

#define RIGHT(code, text, fGeneral, fSpecific)     \
   { &m_guidNULL, code, L ## text,                 \
      (0 | (fGeneral ? SI_ACCESS_GENERAL : 0) |    \
           (fSpecific ? SI_ACCESS_SPECIFIC : 0)) }
///////////////////////////////////////////////////////////////////////////////


//#ifdef SECINFO_IMPL
//Comment out by zzf

///////////////////////////////////////////////////////////////////////////////





///////////////////////////////////////////////////////////////////////////////


#pragma comment(lib, "ACLUI.lib")   // Force linking against this library


///////////////////////////////////////////////////////////////////////////////


//#endif   // SECINFO_IMPL

///////////////////////////////// End of File /////////////////////////////////
#endif